body {
  width: 800px;
  font-family: sans-serif;
  margin: 0 auto;
}

pre.sourceCode, code {
  background-color: #eee;
}

.controls {
  text-align: center;
}

a {
  text-decoration: none;
  color: #888;
  display: inline-block;
}

a:hover {
  text-decoration: underline;
  color: #666;
}

.boxes {
  width: 100%;
  margin-bottom: 1em;
  text-align: center;
}

.box {
  text-align: left;
  width: 395px;
  display: inline-block;
  height: 8em;
  border: 1px solid black;
  vertical-align: top;
  background-color: white;
  border-radius: 5px;
}

.box ul {
  margin: 0;
}

.title {
  padding: 5px;
  font-weight: bold;
}

.controls a {
  display: inline-block;
  padding: 2px;
  border: 1px solid #888;
  border-radius: 2px;
  background-color: #fffef7;
  color: #353535;
}

.controls a:hover {
  text-decoration: none;
  border: 1px solid #353535;
  background-color: #fffceb;
}

.caption {
  text-align: center;
  font-style: italic;
  margin: 0;
}

.figure {
  border: 1px solid black;
  padding: 2em;
}
